/*
    Simple CSAA demo

    This sample demonstrates how to create a window with Coverage Sampling
    Antialiasing using the NV_multisample_coverage extension.
    http://developer.nvidia.com/object/coverage-sampled-aa.html

    In general we recommend using framebuffer objects (FBOs) with the
    NV_framebuffer_multisample_* extensions for offscreen rendering with
    AA, but some applications may require creating a main window with CSAA,
    in which case this method is necessary.

    Author: Rouslan Dimitrov, Devtech, NVIDIA Corporation
    Sample based on the original multisample demo by Matt Craighead
    Additional comments by Mark Kilgard, NVIDIA Corporation
    Email: sdkfeedback@nvidia.com
*/

#pragma comment(lib, "opengl32.lib")
#pragma comment(lib, "glew32.lib")

#include <stdio.h>
#include <stdlib.h>
#include <gl/glew.h>
#include <gl/wglew.h>
#include <gl/glu.h>
#include <gl/glut.h>
#include <windows.h>

#ifndef GL_NV_multisample_coverage
#define GL_COVERAGE_SAMPLES_NV            0x80A9
#define GL_COLOR_SAMPLES_NV               0x8E20
#endif

#ifndef WGL_NV_multisample_coverage
#define WGL_COVERAGE_SAMPLES_NV           0x2042
#define WGL_COLOR_SAMPLES_NV              0x20B9
#endif

#define WIDTH                             800
#define HEIGHT                            600

HWND hWnd;
HDC hDC;
HGLRC hGLRC;
int pixelFormat = 0;
int numColorSamples = 8;
int numCoverageSamples = 16;

bool multisample = true;

const static PIXELFORMATDESCRIPTOR pfd = {
    sizeof(PIXELFORMATDESCRIPTOR),
    1,
    PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
    PFD_TYPE_RGBA,
    24,
    0, 0, 0, 0, 0, 0,
    8, 0,
    0, 0, 0, 0, 0,
    24, 8,
    0,
    PFD_MAIN_PLANE,
    0,
    0,
    0,
    0,
};

static void Display(void)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    if (multisample) {
        glEnable(GL_MULTISAMPLE_ARB);
    } else {
        glDisable(GL_MULTISAMPLE_ARB);
    }

    glBegin(GL_QUADS);
    glColor4f(1,0,0,0);
    glVertex2f(-0.8f, -0.9f);
    glColor4f(0,1,0,1);
    glVertex2f(0.9f, -0.8f);
    glColor4f(0,0,1,0);
    glVertex2f(0.8f, 0.9f);
    glColor4f(1,1,0,1);
    glVertex2f(-0.9f, 0.8f);
    glEnd();

	glutSolidTeapot(0.5);

    SwapBuffers(hDC);
}

void KeyboardFunc(char key)
{
    switch(key) {
    case 'm':
        multisample = !multisample;
        break;
    }
}

static LRESULT CALLBACK WindowProc(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{
    int iAttributes[16];
    FLOAT fAttributes[] = {0, 0};
    UINT numFormats;
    RECT ClientRect;
    BOOL status;
	static int realWindow;

    switch (Msg) {
    case WM_CREATE:
		hDC = GetDC(hWnd);

		// is this is the init window?
		if(!realWindow){
			pixelFormat = ChoosePixelFormat(hDC, &pfd);
			SetPixelFormat(hDC, pixelFormat, &pfd);
			hGLRC = wglCreateContext(hDC);
			wglMakeCurrent(hDC, hGLRC);

			if(glewInit() != GLEW_OK) {
				fprintf(stderr, "main: Couldn't init glew\n");
				Sleep(2000);
				exit(0);
			}
			break;
		}

        // Choose a Pixel Format Descriptor (PFD) with multisampling support.
        iAttributes[0] = WGL_SAMPLE_BUFFERS_ARB;
        iAttributes[1] = 1;
        iAttributes[2] = WGL_COLOR_SAMPLES_NV;
		iAttributes[3] = numColorSamples;
		iAttributes[4] = WGL_COVERAGE_SAMPLES_NV;
		iAttributes[5] = numCoverageSamples;

        iAttributes[6] = WGL_DOUBLE_BUFFER_ARB;
        iAttributes[7] = 1;
        iAttributes[8] = 0;
        iAttributes[9] = 0;

        status = wglChoosePixelFormatARB(hDC, iAttributes, fAttributes, 1, &pixelFormat, &numFormats);
        if (status == FALSE) {
            fprintf(stderr, "main: Couldn't choose a multisample pixel format descriptor\n");
            Sleep(2000);
            exit(0);
        }
        status = SetPixelFormat(hDC, pixelFormat, &pfd);
        if (status == FALSE) {
            fprintf(stderr, "main: Couldn't SetPixelFormat to a multisample pixel format descriptor\n");
            Sleep(2000);
            exit(0);
        }
        hGLRC = wglCreateContext(hDC);
        wglMakeCurrent(hDC, hGLRC);

        // query number of samples
        GLint samples, colorSamples, coverageSamples;
        glGetIntegerv(GL_SAMPLES_ARB, &samples);
        glGetIntegerv(GL_COLOR_SAMPLES_NV, &colorSamples);
        glGetIntegerv(GL_COVERAGE_SAMPLES_NV, &coverageSamples);
        printf("samples: %d, color samples: %d, coverage samples: %d\n", samples, colorSamples, coverageSamples);
        break;

    case WM_DESTROY:
        wglMakeCurrent(NULL, NULL);
        wglDeleteContext(hGLRC);
        ReleaseDC(hWnd, hDC);
		if(realWindow)
            PostQuitMessage(0);
		realWindow = 1;
        break;

    case WM_CHAR:
        KeyboardFunc((char) wParam);
        InvalidateRect(hWnd, NULL, FALSE);
        break;

    case WM_KEYDOWN:
        switch (wParam) {
        case VK_ESCAPE:
            DestroyWindow(hWnd);
            return 0;
        }
        break;
    case WM_SIZE:
	    GetClientRect(hWnd, &ClientRect);
        glViewport(0, 0, ClientRect.right, ClientRect.bottom);
        break;
    case WM_PAINT:
        Display();
        break;
    }
    return DefWindowProc(hWnd, Msg, wParam, lParam);
}

void CheckSupport() {
	// Check for support for CSAA
	if(!glutExtensionSupported("GL_NV_multisample_coverage")) {
        fprintf(stderr, "main: This sample requires the GL_NV_multisample_coverage extensions\n");
        Sleep(2000);
        exit(0);
    }

    int attrib = WGL_NUMBER_PIXEL_FORMATS_ARB;
    int numFormats = 0;
    if (!wglGetPixelFormatAttribivARB(hDC, 1, 0, 1, &attrib, &numFormats)) {
        printf("Error getting number of pixel formats\n");
    }

	int attribs[] = {
        WGL_SAMPLES_ARB,
        WGL_COLOR_SAMPLES_NV,
        WGL_COVERAGE_SAMPLES_NV
    };
	int iValues[3];
    for(int i=1; i<=numFormats; i++){
		wglGetPixelFormatAttribivARB(hDC, i, 0, 3, attribs, iValues);
        printf("pf %d: samples %d, color %d, coverage %d\n", i, iValues[0], iValues[1], iValues[2]);
	}
}

int main(int argc, char **argv)
{
	HINSTANCE hInstance;
    MSG msg;
	WNDCLASS wc;
	char windowTitle[128];

	if(argc != 3){
        printf("Usage: %s [num_color_samples] [num_coverage_samples]\n", argv[0]);
	}else{
		numColorSamples = atoi(argv[1]);
		numCoverageSamples = atoi(argv[2]);
	}
    printf("Press 'm' to toggle multisampling\n");

    hInstance = GetModuleHandle(NULL);

    wc.style         = CS_OWNDC;
    wc.lpfnWndProc   = WindowProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = LoadIcon(NULL, IDI_WINLOGO);
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = "CSAA";

    RegisterClass(&wc);

	// Create temp window to init glew
	hWnd = CreateWindow("CSAA", "Temp Window",
        WS_OVERLAPPEDWINDOW,
        40, 40,
        WIDTH,
        HEIGHT,
        NULL, NULL, hInstance, NULL);
    if (!hWnd) {
        fprintf(stderr, "main: Couldn't create init window\n");
        Sleep(2000);
        return 0;
    }

	CheckSupport();

    // Now destroy the fake window; we only need it to initialize WGL extensions.
    DestroyWindow(hWnd);

	sprintf(windowTitle, "OpenGL CSAA Sample: %d color samples, %d coverage samples", numColorSamples, numCoverageSamples);

    // Create the "real" multisampled window.
    hWnd = CreateWindow("CSAA", windowTitle,
        WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
        40, 40,
        WIDTH,
        HEIGHT,
        NULL, NULL, hInstance, NULL);
    if (!hWnd) {
        fprintf(stderr, "main: Couldn't create window\n");
        Sleep(2000);
        return 0;
    }
    ShowWindow(hWnd, SW_SHOW);
    UpdateWindow(hWnd);
    SetFocus(hWnd);

    while (GetMessage(&msg, NULL, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    return msg.wParam;
}
